# Create XML to invoke a SOAP method

toSOAP <-
function(obj, con = xmlOutputBuffer(header=""))
{

 writeTypes <- function(x) {
      types <- getSOAPType(x)
      if(!is.null(types)) {
       for(a in names(types)) {
         cat(a,"=\"", types[[a]],"\"", sep="", file=con)
       }
      }   
 }
  
 # con$add("value")
 if(length(obj) > 1) {
  if(!is.null(names(obj))) {
    for(i in names(obj)) {
      cat("<", i, file = con, sep="")
      writeTypes(obj[[i]])
      cat(obj[[i]])
    }
  } else {
    for(i in obj) {
      cat("<item ", file=con)
      writeTypes(i)
      cat(">\r\n", file=con)
      cat(i, file=con)
      cat("\r\n</item>", file=con)      
    }
  }
 } else {
    cat(obj, file=con)
 }

 invisible(TRUE)
}  

SOAPTypes <-
  list("character" = c("xsi:type" = "xsd:string"),
       "numeric" = c("xsi:type" = "xsd:float"),
       "double" = c("xsi:type" = "xsd:float"),       
       "integer" = c("xsi:type" = "xsd:int"),
       "logical" = c("xsi:type" = "xsd:boolean"),
       "NULL" = c("xsi:null" = 1)
    )

getSOAPType <-
function(obj)
{
  if(is.null(obj) || length(obj) == 0)
    return(SOAPTypes[["NULL"]])

  if(length(obj) > 1) {
    # Array
    n <- length(obj)
    same <- sapply(obj, function(x, target) typeof(x) == target, target = typeof(obj[[1]]))

    ans <- c("SOAP-ENC:Array")
    if(all(same)) {
      type <- getSOAPType(obj[[1]])
      ans[["SOAP-ENC:arrayType"]]  = paste(type[1], "[", n, "]", sep="")
    }
    
  } else {
     m <- typeof(obj)
     ans <- SOAPTypes[[m]]
  }
  
  return(ans)  
}  

.SOAPAction <-
function(action, method, server, xmlns)
{
  paste(action, "#", method, sep="")
}

.SOAPDefaultHandlers <-
  list(action = .SOAPAction,
       result = function(xmlSource, header, method, server) {
         response <- parseSOAP(xmlSource, asText = TRUE)
         fromSOAP(response[[1]])
       })

SOAPHandlers <-
function(..., include = character(0), exclude = character(0))
{  
 defaults <- .SOAPDefaultHandlers
 els <- list(...)

 .merge(els, defaults, include, exclude)
}

.SOAP <-
function(server, method, ..., action, nameSpaces = SOAPNameSpaces(), xmlns = NULL,
          handlers = SOAPHandlers(), debug = FALSE)  
{
  # Get the connection.
 if(debug) {
   con <- textConnection(".SOAPTest", "w")
 } else {
   con <- socketConnection(server@host, port = server@port, open = "w+", blocking = TRUE)
 }
 on.exit(close(con))

 if(is.null(xmlns))
   xmlns <- c(namesp1 = action)

 if(!is.null(handlers) && !is.na(match("action", names(handlers)))) 
   action <- handlers[["action"]](action, method, server, xmlns)
 
 writeSOAPHeader(server@url, server@host, action, con)

 txt <- writeSOAPMessage(con, nameSpaces, method, ..., xmlns = xmlns)

 cat("Content-Length: ", nchar(txt),"\r\n", sep="", file=con)
 cat("\r\n", file=con)
 cat(txt, sep="\n", file=con)

 cat("\n", file=con)
 
 if(debug) {
   return(.SOAPTest)
 }


  # Now read the response from the server.
 txt <- readLines(con) 

  # Parse this into the HTTP header and the payload. 
 content <- getHTTPContent(txt)

  # Check things were ok with the HTTP.
 if(isHTTPError(content$header)) {
     # This would be the exception that we throw if we had an exception system.
   fault <- SOAPFault(parseSOAP(content$content, asText = TRUE))
    # 
   stop("Error occurred in the HTTP request", fault@message, fault@detail)
   return(fault)
 }

 if(!is.null(handlers) && !is.na(match("result", names(handlers))))
   return(handlers[["result"]](content$content,  content$header, method))

 return(content)
}


writeSOAPHeader <-
function(url, host, action, con, ...)
{
 cat("POST /", gsub("^/", "", url), " HTTP/1.1\r\n", sep="", file=con)
 cat("Host: ", host, "\r\n", sep="", file=con)
 cat("Content-Type: text/xml; charset=utf-8\r\n", file=con)
 if(!is.null(action))
   cat("SOAPAction: \"", action, "\"\r\n", sep="", file=con)

 args <- list(...)
 for(i in names(args)) {
    cat(i, ": ", args[[i]], sep="", file=con)
 }

}  


writeSOAPMessage <-
function(con, nameSpaces, method, ..., xmlns = NULL)
{
 con <- textConnection(".foo", open="w")
 writeSOAPEnvelope(con, nameSpaces = nameSpaces)

 writeSOAPBody(method, ..., xmlns = xmlns, con = con)

 cat("</SOAP-ENV:Envelope>\r\n", file=con)

  # Flushing to make certain the commands go to the server.
 cat("\r\n", file=con)
 close(con)
 paste(get(".foo"), collapse="\n")
}  


#SOAP 1.1
#      "SOAP-ENC" = 'http://schemas.xmlsoap.org/soap/encoding/',
#      "xsd" = 'http://www.w3.org/1999/XMLSchema',
# SOAP  1.2 
#      "SOAP-ENV" = 'http://www.w3.org/2001/06/soap-envelope',
#      "SOAP-ENC" = 'http://www.w3.org/2001/06/soap-encoding',
#      "xsd" = 'http://www.w3.org/2001/XMLSchema',

.SOAPDefaultNameSpaces <-
  list("1.1" = 
         c(
           'SOAP-ENC'="http://schemas.xmlsoap.org/soap/encoding/",
           'SOAP-ENV'="http://schemas.xmlsoap.org/soap/envelope/",           
           'xsi'="http://www.w3.org/1999/XMLSchema-instance",
           'xsd'="http://www.w3.org/1999/XMLSchema"
           )  ,
       "1.2" = 
         c(
           'SOAP-ENC'="http://www.w3.org/soap-encoding/",
           'SOAP-ENV'="http://www.w3.org/2001/06/soap-envelope",
           'xsi'="http://www.w3.org/2001/XMLSchema-instance",
           'xsd'="http://www.w3.org/2001/XMLSchema"
           )  )       
       

.merge <-
function(els, defaults, include = NULL, exclude = NULL)
{
 if(length(els) > 0) {
   which <- match(names(defaults), names(els))
   if(any(!is.na(which)))
     els[names(defaults)[!is.na(which)]] <- defaults[!is.na(which)]
 } else
   els <- defaults

 if(length(include)) {
   els <- els[include]
 } else if(length(exclude)) {
   which <- match(exclude, names(els))
   if(any(!is.na(which)))
     els <- els[- (which[!is.na(which)])]
 }

 els   
}  

SOAPNameSpaces <-
function(..., include = character(0), exclude = character(0), version = "1.1")
{
  defaults <- .SOAPDefaultNameSpaces[[version]]
  els <- sapply(list(...), as.character)
  .merge(els, defaults, include, exclude)
}

  # This and writeSOAPBody should use the XMLOutput.. classes in the XML package.
writeSOAPEnvelope <-
function(con, nameSpaces = .SOAPNameSpaces())
{
  cat("<?xml version='1.0' encoding='UTF-8'?>\r\n", file = con)
  cat("<SOAP-ENV:Envelope\r\n", file=con)
  for(i in names(nameSpaces)) {
     cat("xmlns:", i, "=\"", nameSpaces[i],"\"\r\n", sep="", file=con)
  }
  cat('SOAP-ENV:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/"', file=con)
  cat(">\r\n", file=con)
}


writeSOAPBody <-

function(method, ..., xmlns= NULL, con)
{
 cat('<SOAP-ENV:Body>\r\n', file=con)

 nsName <- ""
 if(!is.null(xmlns) && !is.null(names(xmlns))) {
   nsName <- paste(names(xmlns)[1], ":", sep="")
 }
 
 cat("<", nsName, method, sep="", file=con)  

 if(!is.null(xmlns) && xmlns != "") {
   ns <- ""
   if(!is.null(names(xmlns)))
     ns <-  paste(":", names(xmlns)[1], sep="")
   cat(" xmlns", ns,"=\"", xmlns,"\"", file=con, sep="")
 }
 cat(">", file=con)

 args <- list(...)
 argNames <- names(args)

 if(length(args) > 0) {
  for(i in 1:length(args)) {
   
    type <- getSOAPType(args[[i]])
    if(!is.null(argNames))
      argName <- argNames[i]
    else
      argName <- letters[i]

    cat("<", argName, sep="", file=con)
    if(type != "") 
      cat(" xsi:type=\"", type, "\"", file=con, sep="")
    cat(">", file=con)
    toSOAP(args[[i]], con)
    cat("</", argName, ">\r\n", sep="", file=con)  
  }
 }
 cat("</", nsName, method, ">\r\n", sep="", file=con) 

 cat("</SOAP-ENV:Body>\r\n", file=con) 
}  




